from typing import List

from flask import current_app
from sqlalchemy import func, and_, or_, false

from whatis.models import Whatis
from whatis.proxies import db_session


def postgres_lookup(input: str) -> List[Whatis]:
    subquery_base = db_session.query(
        Whatis.whatis_id, func.max(Whatis.version).label("version")
    )
    subquery_filtered = subquery_base.filter(
        or_(
            func.levenshtein(func.lower(Whatis.terminology), input.lower()) <= 1,
            Whatis.terminology.ilike(f"%{input}%") if len(input) > 3 else false(),
        )
    )
    subquery_grouped = subquery_filtered.group_by(Whatis.whatis_id).subquery("s2")
    query = db_session.query(Whatis).join(
        subquery_grouped,
        and_(
            Whatis.whatis_id == subquery_grouped.c.whatis_id,
            Whatis.version == subquery_grouped.c.version,
        ),
    )
    return query.all()


def sqlite_lookup(input: str) -> List[Whatis]:
    subquery_base = db_session.query(
        Whatis.whatis_id, func.max(Whatis.version).label("version")
    )
    subquery_filtered = subquery_base.filter(
        Whatis.terminology.ilike(f"%{input}%" if len(input) > 3 else f"{input}")
    )
    subquery_grouped = subquery_filtered.group_by(Whatis.whatis_id).subquery("s2")
    query = db_session.query(Whatis).join(
        subquery_grouped,
        and_(
            Whatis.whatis_id == subquery_grouped.c.whatis_id,
            Whatis.version == subquery_grouped.c.version,
        ),
    )
    return query.all()


def lookup_whatis(input: str) -> List[Whatis]:
    dialect = current_app.config.get("DB_DIALECT")

    if dialect == "postgres":
        current_app.logger.debug(f"Using postgres to lookup {input}")
        return postgres_lookup(input)
    elif dialect == "sqlite":
        current_app.logger.debug(f"Using sqlite to lookup {input}")
        return sqlite_lookup(input)
    else:
        raise RuntimeError(
            f"Unrecognised SQL dialect {dialect} - I don't even know how you got here!"
        )


def get_all_whatises() -> List[Whatis]:
    # Return all Whatises to be sent as a CSV
    subquery_base = db_session.query(
        Whatis.whatis_id, func.max(Whatis.version).label("version")
    )
    subquery_grouped = subquery_base.group_by(Whatis.whatis_id).subquery("s2")
    query = (
        db_session.query(Whatis)
        .join(
            subquery_grouped,
            and_(
                Whatis.whatis_id == subquery_grouped.c.whatis_id,
                Whatis.version == subquery_grouped.c.version,
            ),
        )
        .order_by(Whatis.terminology)
    )
    return query.all()



import os
from functools import partial

from flask.cli import FlaskGroup

from whatis.app import WhatisApp
from whatis.config import WhatisConfig
from whatis.models import Whatis


app = WhatisApp(config=WhatisConfig, DB_AUTO_CREATE=False, DB_AUTO_UPGRADE=False)
cli = FlaskGroup(create_app=partial(WhatisApp, config=WhatisConfig))


@cli.command("recreate_db")
def recreate_db():
    app.db.drop_all()
    app.db.create_all()
    app.db.session.commit()


@cli.command("seed_db")
def seed_db():
    """Seeds the database."""
    USER_ID = "UX8FHM5"
    models = [
        Whatis(
            terminology="whatis_a",
            definition="The definition of whatis A",
            notes="Some more notes about Whatis A",
            links="www.google.com",
            version=1,
            added_by=USER_ID,
            point_of_contact="U9KR5QZA5",
        ),
        Whatis(
            terminology="whatis_b",
            definition="The definition of whatis B",
            notes="Some more notes about Whatis B",
            links="www.facebook.com",
            version=1,
            added_by=USER_ID,
            point_of_contact="U9KR5QZA5",
        ),
    ]
    for m in models:
        app.db.session.add(m)
    app.db.session.commit()


if __name__ == "__main__":
    cli()



import json
import os
from json import loads, dumps
from unittest.mock import MagicMock, patch
from flask import Response

import pytest

from whatis import constants
from whatis.app import WhatisApp
from whatis.config import WhatisConfig
from whatis.models import Whatis
from whatis.utils.dialog_components import (
    TERMINOLOGY_KEY,
    DEFINITION_KEY,
    NOTES_KEY,
    LINKS_KEY,
    POINT_OF_CONTACT_KEY,
)

TEST_WHATIS = Whatis(
    id=2,
    whatis_id="sfCw",
    terminology="UDF",
    definition="User defined function",
    notes="A common feature to many SQL languages",
    links="google,facebook",
    version=0,
    added_by="UX8HGN",
    point_of_contact="C5TY9IL",
)


class TestingConfig(WhatisConfig):
    SQLALCHEMY_DATABASE_URI = "sqlite:///:memory:"
    SLACK_SIGNING_SECRET = "secrety secret"
    SLACK_TOKEN = "tokeny token"
    DEBUG = True


def create_slash_command_input(text, user_id="U9KR5QZA5"):
    return dict(
        text=text,
        response_url="https://slack.webhook.com/TN04R",
        user_id=user_id,
        user_name="Steff Nezos",
        channel_id="C9Z2KJTTB",
        channel_name="cat-factoids",
        trigger_id="1032423.342",
        command="/whatis",
        team_id="C4RRSV",
        team_domain="parragon",
    )


def create_block_action(action_id, block_id, action_value, user_id="U9KR5QZA5"):
    return dict(
        payload=dumps(
            dict(
                user=dict(id="U9KR5QZA5", username="some dude"),
                response_url="https://slack.webhook.com/TN04R",
                channel=dict(id="CX6TY", name="some channel"),
                trigger_id="1032423.342",
                command="/whatis",
                team=dict(id="C4RRSV", domain="team mcteam"),
                type="block_actions",
                container=dict(message_ts="2242234.234"),
                actions=[
                    dict(action_id=action_id, block_id=block_id, value=action_value)
                ],
            )
        )
    )


def create_dialog_submit_action(callback_id):
    return dict(
        payload=dumps(
            dict(
                response_url="https://slack.webhook.com/TN04R",
                type="dialog_submission",
                user=dict(id="U9KR5QZA5", name="some dude"),
                channel=dict(id="CX6TY", name="some channel"),
                team=dict(id="C4RRSV", domain="team mcteam"),
                callback_id=callback_id,
                state="",
                submission={
                    POINT_OF_CONTACT_KEY: TEST_WHATIS.added_by,
                    LINKS_KEY: TEST_WHATIS.links,
                    NOTES_KEY: TEST_WHATIS.notes,
                    TERMINOLOGY_KEY: TEST_WHATIS.terminology,
                    DEFINITION_KEY: TEST_WHATIS.definition,
                },
            )
        )
    )


@pytest.fixture
def client():
    client = WhatisApp(config=TestingConfig)
    client.sc = MagicMock()
    with client.app_context():
        client.db.session.add(
            Whatis(
                terminology="eod",
                definition="End of Day",
                notes="Used when you dont have time to say end of day",
                links="google,facebook",
                point_of_contact="C9Z2KJEVB",
                version=0,
                added_by="U9KR5QZA5",
            )
        )
        client.db.session.add(
            Whatis(
                terminology="Lost customer",
                whatis_id="wE5Bn",
                definition="When a customer does not make any repeat transactions for > 20 days",
                notes="Decided on by Data science team",
                links=None,
                point_of_contact=None,
                version=0,
                added_by="U9KR5QZB4",
            )
        )
        client.db.session.add(
            Whatis(
                terminology="Lost customer",
                whatis_id="wE5Bn",
                definition="When a customer does not make any repeat transactions for > 20 days",
                notes="Decided on by Data science team and CIO",
                links="jira",
                point_of_contact=None,
                version=1,
                added_by="U9KR5QZA5",
            )
        )
        client.db.session.commit()
    return client.test_client()


def test_client_dialect(client):
    assert client.application.config["DB_DIALECT"] == "sqlite"


def test_get_whatis(client):
    resp = client.post("/slack/whatis", data=create_slash_command_input(text="eod"))
    assert len(loads(resp.data)["blocks"]) == 7
    resp = client.post(
        "/slack/whatis", data=create_slash_command_input(text="Lost customer")
    )
    assert len(loads(resp.data)["blocks"]) == 7


@patch("requests.post")
def test_create_new(rp, client):
    # Press the Add new whatis button
    resp = client.post(
        "/slack/actions",
        data=create_block_action(
            action_id=constants.CREATE_NEW_WHATIS_ID,
            block_id="blocky",
            action_value=constants.CREATE_NEW_WHATIS_ID,
        ),
    )
    assert resp.data.decode() == ""

    # Now submit a dialog
    resp = client.post(
        "/slack/actions",
        data=create_dialog_submit_action(
            callback_id=constants.CREATE_NEW_WHATIS_ID_SUBMIT
        ),
        content_type="application/x-www-form-urlencoded",
    )
    assert resp.data.decode() == ""
    rp.assert_called()

    # Now query that same whatis we just created and ensure it returns result
    resp = client.post(
        "/slack/whatis", data=create_slash_command_input(text=TEST_WHATIS.terminology)
    )
    assert len(loads(resp.data)["blocks"]) == 7


@pytest.fixture()
def preload_filepath():
    fpath = "_testing_whatis_preload.json"
    with open(fpath, "w") as file:
        file.write(
            json.dumps(
                [
                    dict(terminology="eod", definition="end of day"),
                    dict(
                        terminology="FBI",
                        definition="Federal Bureau of Intelligence",
                        links="https://www.jira.com/issues/DE-356",
                    ),
                    dict(
                        terminology="CSA",
                        definition="Corporate Social allowance",
                        notes="Here are some notes",
                    ),
                ]
            )
        )
    yield fpath
    os.remove(fpath)


@pytest.fixture()
def preload_filepath_broken():
    fpath = "_testing_whatis_preload_broken.json"
    with open(fpath, "w") as file:
        file.write(
            json.dumps(
                [
                    dict(terminology="eod", definition="end of day"),
                    dict(
                        terminology="FBI",
                        definition="Federal Bureau of Intelligence",
                        links="https://www.jira.com/issues/DE-356",
                    ),
                    dict(
                        fasdfa="CSA",
                        definition="Corporate Social allowance",
                        notes="Here are some notes",
                    ),
                ]
            )
        )
    yield fpath
    os.remove(fpath)


def test_preload_whatises(preload_filepath, preload_filepath_broken):
    client = WhatisApp(config=TestingConfig, preload_path=preload_filepath)

    tc = client.test_client()
    resp = tc.post("/slack/whatis", data=create_slash_command_input(text="fbi"))
    assert len(loads(resp.data)["blocks"]) == 7

    client.handle_whatis_preload(preload_filepath)

    # Check that loads don't happen again
    resp = tc.post("/slack/whatis", data=create_slash_command_input(text="fbi"))
    assert len(loads(resp.data)["blocks"]) == 7

    # Test runtime error with missing filepaths
    with pytest.raises(FileNotFoundError):
        client = WhatisApp(config=TestingConfig, preload_path="Idontexist.json")

    # Test corrupted filepaths
    with pytest.raises(RuntimeError):
        client = WhatisApp(config=TestingConfig, preload_path=preload_filepath_broken)


def test_load_from_args(preload_filepath):
    class args:
        db = "sqlite:///:memory:"
        slack_token = "token"
        slack_signing_secret = "sdfasd"
        debug = False
        admin_user_ids = "UX3G1N,UX5G1T"
        admin_channel_ids = "C67UOP"

    a = args()
    a.preload_filepath = preload_filepath

    client = WhatisApp(
        config=WhatisConfig.from_args(a), preload_path=a.preload_filepath
    )
    assert client.config.get("SQLALCHEMY_DATABASE_URI") == args.db
    assert client.config.get("SLACK_TOKEN") == args.slack_token
    assert client.config.get("SLACK_SIGNING_SECRET") == args.slack_signing_secret
    assert client.config.get("DEBUG") == args.debug
    assert client.config.get("ADMIN_CHANNEL_IDS") == args.admin_channel_ids.split(",")
    assert client.config.get("ADMIN_USER_IDS") == args.admin_user_ids.split(",")


def test_slack_verification(client):
    client.application.config["DEBUG"] = False
    resp: Response = client.post(
        "/slack/whatis", data=create_slash_command_input(text="fbi")
    )

    # Make sure it is denying calls with no headers
    assert resp.status_code == 403
    assert "Request signature verification failed" in resp.data.decode()

    # Make sure our verification also fails when we have correct verification headers but wrong signing secret
    resp: Response = client.post(
        "/slack/whatis",
        data=create_slash_command_input(text="fbi"),
        headers={
            "X-Slack-Request-Timestamp": "1570368107",
            "X-Slack-Signature": "v0=70005dbf3ebd0cd689fbc149bad4b21280b2476bbf6d9745c49e305798080bec",
        },
    )

    assert resp.status_code == 403
    assert "Request signature verification failed" in resp.data.decode()



from flask import current_app
from slack.web.client import WebClient
from sqlalchemy.orm.session import Session
from werkzeug.local import LocalProxy

__all__ = ["db_session", "slack_client"]

db_session: Session = LocalProxy(lambda: current_app.db.session)
slack_client: WebClient = LocalProxy(lambda: current_app.sc)



import logging
import os
import typing

from dotenv import find_dotenv, load_dotenv

load_dotenv(find_dotenv())
logger = logging.getLogger(__name__)


class WhatisConfig:

    SQLALCHEMY_TRACK_MODIFICATIONS: bool = False
    ADMIN_USER_IDS: typing.List[str] = []
    ADMIN_CHANNEL_IDS: typing.List[str] = []
    SLACK_TOKEN: typing.Optional[str] = os.environ.get("SLACK_TOKEN")
    SLACK_SIGNING_SECRET: typing.Optional[str] = os.environ.get("SLACK_SIGNING_SECRET")
    SQLALCHEMY_DATABASE_URI: str = "sqlite:///:memory:"

    @classmethod
    def from_args(cls, args):

        if args.db is not None:
            cls.SQLALCHEMY_DATABASE_URI = args.db

        if args.debug is not None:
            cls.DEBUG = args.debug

        if args.admin_user_ids is not None:
            cls.ADMIN_USER_IDS = args.admin_user_ids.split(",")

        if args.admin_channel_ids is not None:
            cls.ADMIN_CHANNEL_IDS = args.admin_channel_ids.split(",")

        if args.slack_token is not None:
            cls.SLACK_TOKEN = args.slack_token

        if args.slack_signing_secret is not None:
            cls.SLACK_SIGNING_SECRET = args.slack_signing_secret

        return cls()



from flask import current_app
from sqlalchemy import desc
from io import StringIO

from whatis.models import Whatis
from whatis.proxies import db_session, slack_client
from whatis.utils import dialog_components, responder
from whatis.utils.message_components import build_channel_whatis
from whatis.utils.lookups import get_all_whatises
from whatis import constants


def send_create_form(trigger_id):
    slack_client.dialog_open(
        trigger_id=trigger_id,
        dialog=dialog_components.build_create_new_dialog().to_dict(),
    )
    return "", 204


def send_update_form(trigger_id, whatis_id: int):
    slack_client.dialog_open(
        trigger_id=trigger_id,
        dialog=dialog_components.build_update_dialog(
            whatis=get_whatis(whatis_id)
        ).to_dict(),
    )
    return "", 204


def create_whatis(
    terminology,
    definition,
    whatis_id=None,
    notes=None,
    links=None,
    version=None,
    added_by=None,
    point_of_contact=None,
) -> Whatis:
    whatis = Whatis(
        terminology=terminology,
        definition=definition,
        whatis_id=whatis_id,
        notes=notes,
        version=version,
        links=links,
        added_by=added_by,
        point_of_contact=point_of_contact,
    )
    db_session.add(whatis)
    db_session.commit()
    current_app.logger.info(f"Created the Whatis: {whatis}")
    return whatis


def update_whatis(id, definition, notes, links, added_by, point_of_contact) -> Whatis:
    wi: Whatis = db_session.query(Whatis).filter(Whatis.id == id).first()
    # Check to see if a different user updated your terminology
    if added_by != wi.added_by:
        slack_client.chat_postMessage(
            channel=wi.added_by,
            text=f"Hey  <@{wi.added_by}>, the user <@{added_by}> has updated your Whatis for *{wi.terminology}*, "
            f"try searching for it again to see what they changed!",
        )
    create_whatis(
        terminology=wi.terminology,
        whatis_id=wi.whatis_id,
        definition=definition,
        notes=notes,
        links=links,
        version=wi.version + 1,
        added_by=added_by,
        point_of_contact=point_of_contact,
    )
    return wi


def delete_whatis(whatis_id):
    wi = db_session.query(Whatis).filter(Whatis.whatis_id == whatis_id).all()
    for w in wi:
        db_session.delete(w)
    db_session.commit()


def rollback_whatis(id):
    wi = (
        db_session.query(Whatis)
        .filter(Whatis.id == id)
        .order_by(desc(Whatis.version))
        .first()
    )
    db_session.delete(wi)
    db_session.commit()


def get_whatis(id: int) -> Whatis:
    wi = db_session.query(Whatis).filter(Whatis.id == id).first()
    return wi


def send_all_tsv(user_id: str):
    all_whatis = get_all_whatises()
    whatis_tsv = "\t".join([i.capitalize() for i in constants.WHATIS_FIELDS])
    for wi in all_whatis:
        whatis_tsv += "\n" + "\t".join(
            getattr(wi, a) or "" for a in constants.WHATIS_FIELDS
        )
    slack_client.files_upload(
        content=whatis_tsv, channels=user_id, filename="all_company_terminology.tsv"
    )
    return "A list of all Whatises for your company is on its way to you :fire_engine:"


def send_to_channel(channel_id: str, id: int, user: str):
    wi = get_whatis(id)
    message = build_channel_whatis(triggering_user=user, whatis=wi)
    slack_client.chat_postMessage(
        channel=channel_id,
        blocks=[i.to_dict() for i in message.blocks],
        text=message.text,
    )



import json
import logging

from flask import request, Blueprint, current_app, abort, jsonify
from slack.web.classes.interactions import SlashCommandInteractiveEvent

from whatis.routes.actions import send_create_form
from whatis.routes.block_action_handlers import handle_block_actions
from whatis.routes.dialog_handlers import handle_dialog_submission
from whatis.utils.lookups import lookup_whatis
from whatis.utils.message_components import build_whatis_message
from whatis.utils.request import verify_slack_request

slack_blueprint = Blueprint(__name__, "slack")

logger = logging.getLogger(__name__)


@slack_blueprint.before_request
def verify_request():
    verify_slack_request()


@slack_blueprint.route("/whatis", methods=["POST"])
def get_whatis_rule():
    sc = SlashCommandInteractiveEvent(request.form.to_dict())

    terminology = sc.text
    # Shortcut command to create new terminology
    if terminology == "create":
        return send_create_form(sc.trigger_id)

    elif terminology == "":
        return "Oops! you need to give me something to search :pensive:"

    wi = lookup_whatis(terminology)
    current_app.logger.debug(f"Whatis query result for {terminology}: {wi}")
    is_admin = sc.user.id in current_app.admin_users or current_app.admin_users == []
    message = build_whatis_message(terminology, wi, is_admin=is_admin)
    return jsonify(message.to_dict())


@slack_blueprint.route("/actions", methods=["POST"])
def handle_action():
    request_data = request.form.to_dict()
    current_app.logger.debug(request_data)

    action_payload = json.loads(request_data.get("payload"))
    action_type = action_payload.get("type")

    if action_type == "dialog_submission":
        return handle_dialog_submission(action_payload)
    elif action_type == "block_actions":
        return handle_block_actions(action_payload)
    else:
        abort(400, f"Unknown action type {action_type}")
    return "", 204



import logging

from flask import request, abort, current_app
from slack.web.client import WebClient

logger = logging.getLogger(__name__)


def verify_slack_request() -> None:
    """
    This verification function is designed to be placed in a flask before_request handler
    """
    request_data = request.get_data().decode()
    current_app.logger.debug(request.headers)
    current_app.logger.debug(request_data)
    timestamp = request.headers.get("X-Slack-Request-Timestamp", "")
    slack_sig = request.headers.get("X-Slack-Signature", "")
    request_verified = False
    verify_message = ""

    if any([timestamp == "", slack_sig == ""]):
        verify_message += "Slack request missing timestamp or signature headers"
    else:
        request_verified = WebClient.validate_slack_signature(
            signing_secret=current_app.config["SLACK_SIGNING_SECRET"],
            timestamp=timestamp,
            signature=slack_sig,
            data=request_data,
        )

    if request_verified is False:
        if current_app.config["DEBUG"] is True:
            current_app.logger.info(
                f"Request verification: {request_verified}, msg: {verify_message}"
            )
        else:
            verify_message = "Request signature verification failed"
            abort(403, verify_message)



from flask import current_app
from slack.web.classes.interactions import MessageInteractiveEvent

from whatis import constants
from whatis.routes.actions import (
    send_create_form,
    send_update_form,
    send_to_channel,
    delete_whatis,
    rollback_whatis,
    send_all_tsv,
)
from whatis.utils.interaction_handler import SlackInteractionHandler
from whatis.utils.responder import webhook_response, basic_responder_response

block_interactor = SlackInteractionHandler()


def handle_block_actions(data: dict):
    data["message"] = {"ts": data["container"]["message_ts"]}
    ah = MessageInteractiveEvent(data)
    result = block_interactor.interact(ah.action_id, action=ah)
    wr = None
    if result is not None:
        wr = webhook_response(ah.response_url, result)
    current_app.logger.info(
        f"Block action interaction for {ah.action_id} returning {result} - {wr}"
    )
    return "", 204


@block_interactor.interaction(constants.CREATE_NEW_WHATIS_ID)
def create_whatis_action(action: MessageInteractiveEvent):
    send_create_form(action.trigger_id)
    return


@block_interactor.interaction(constants.UPDATE_WHATIS_ID)
def update_whatis_action(action: MessageInteractiveEvent):
    send_update_form(action.trigger_id, action.value)
    return


@block_interactor.interaction(constants.WHATIS_SEND_CHANNEL_ID)
def send_to_channel_whatis_action(action: MessageInteractiveEvent):
    send_to_channel(action.channel.id, action.value, action.user.id)
    return


@block_interactor.interaction(constants.ROLLBACK_WHATIS_ID)
def rollback_whatis_action(action: MessageInteractiveEvent):
    rollback_whatis(action.value)
    return basic_responder_response(
        "Whatis successfully rolled back :rolled_up_newspaper:"
    ).to_dict()


@block_interactor.interaction(constants.DELETE_WHATIS_ID)
def delete_whatis_action(action: MessageInteractiveEvent):
    delete_whatis(action.value)
    return basic_responder_response("Whatis successfully deleted :skull:").to_dict()


@block_interactor.interaction(constants.WHATIS_ALL_ID)
def all_whatis_action(action: MessageInteractiveEvent):
    return send_all_tsv(action.user.id)



"""
Defines the different methods that build the Whatis component visible in Slack
"""
from typing import List, Union

from slack.web.classes import blocks, elements, messages, objects

from whatis import constants
from whatis.models import Whatis


def build_whatis_action_confirm(title: str, text: str) -> objects.ConfirmObject:
    return objects.ConfirmObject(
        title=title, text=text, confirm="`Accept`", deny="Cancel"
    )


def build_whatis_actions(whatis: Whatis, is_admin: bool = False) -> blocks.ActionsBlock:
    """
    Adds the message button components to The Whatis message
    Args:
        whatis: A whatis model object
        is_admin: Is the user this component is being built for a workspace whatis admin

    Returns:

    """
    update_whatis_button = elements.ButtonElement(
        text="Update",
        style="primary",
        action_id=constants.UPDATE_WHATIS_ID,
        value=str(whatis.id),
        confirm=build_whatis_action_confirm(
            "Update Whatis",
            'Are you sure you want to update this Whatis? If you do, you will be listed under "Added by" and the '
            "previous updater will be notified!",
        ),
    )
    rollback_whatis_button = elements.ButtonElement(
        text="Rollback",
        action_id=constants.ROLLBACK_WHATIS_ID,
        value=str(whatis.id),
        confirm=build_whatis_action_confirm(
            "Rollback Whatis",
            "Are you sure you would like to rollback this Whatis to its last version?",
        ),
    )
    delete_whatis_button = elements.ButtonElement(
        text="Delete",
        style="danger",
        action_id=constants.DELETE_WHATIS_ID,
        value=whatis.whatis_id,
        confirm=build_whatis_action_confirm(
            "Delete Whatis",
            "This will permanently delete this Whatis along with its whole version history, "
            "are you sure you want to do this?",
        ),
    )
    channel_whatis_button = elements.ButtonElement(
        text="Send to Channel",
        action_id=constants.WHATIS_SEND_CHANNEL_ID,
        value=str(whatis.id),
        confirm=build_whatis_action_confirm(
            "Send Whatis to channel?",
            "This Action will post this Whatis as visible to everyone in the current channel!",
        ),
    )
    actions = [update_whatis_button, channel_whatis_button]
    if is_admin is True:
        actions.extend([rollback_whatis_button, delete_whatis_button])
    return blocks.ActionsBlock(elements=actions)


def build_whatis_component(
    whatis: Whatis, is_admin: bool = False, include_buttons: bool = True
) -> List[Union[blocks.DividerBlock, blocks.SectionBlock, blocks.ActionsBlock]]:
    """
    Build the component that will actually hold the whatis including actions relating to it

    Args:
        whatis: A whatis model object
        is_admin: Is the user this component is being built for a workspace whatis admin
        include_buttons: Should action buttons be included - used for channel messages

    Returns: The component for each Whatis

    """
    whatis_fields = []
    for field in constants.WHATIS_FIELDS:
        # My god I want walrus operators already
        value = getattr(whatis, field)
        if value is not None:
            if field == "links":
                value = "- " + "\n- ".join(value.split(","))
            elif field in ["added_by", "point_of_contact"]:
                mention_char = "@" if value.startswith("U") else "#"
                value = f"<{mention_char}{value}>"
            whatis_fields.append(f"*{field.capitalize().replace('_',' ')}*\n{value}\n")
    component_blocks = [blocks.SectionBlock(fields=whatis_fields)]

    # Add whatis action buttons
    if include_buttons is True:
        component_blocks.append(build_whatis_actions(whatis, is_admin))

    # Add the rest of the block content
    component_blocks.extend(
        [
            # Add whatis contextual information
            blocks.ContextBlock(
                elements=[
                    objects.MarkdownTextObject(
                        text=f"Last updated: {whatis.submitted_at.strftime('%Y-%m-%d')}"
                    ),
                    objects.MarkdownTextObject(text=f"{str(whatis.version)} revisions"),
                ]
            ),
            blocks.DividerBlock(),
        ]
    )
    return component_blocks


def build_whatis_footer(is_admin: bool = False) -> blocks.ActionsBlock:
    """
    Builds the footer to attach to all whatis requests

    Args:
        is_admin: Is the user this component is being built for a workspace whatis admin

    Returns:

    """
    add_new_button = elements.ButtonElement(
        text="Add a new Whatis!",
        style="primary",
        action_id=constants.CREATE_NEW_WHATIS_ID,
        value=constants.CREATE_NEW_WHATIS_ID,
    )
    send_all_button = elements.ButtonElement(
        text="See all Whatises",
        action_id=constants.WHATIS_ALL_ID,
        value=constants.WHATIS_ALL_ID,
        confirm=build_whatis_action_confirm(
            "Get all Whatises?",
            "This action will cause me to send you a TSV of all "
            "your organisation's terminology, this may cause "
            "things to break or take some time.",
        ),
    )
    return blocks.ActionsBlock(elements=[add_new_button, send_all_button])


def build_whatis_message(
    original_query: str, whatises: List[Whatis], is_admin: bool = False
) -> messages.Message:
    block_list = []
    if whatises:
        block_list.extend(
            [
                blocks.SectionBlock(
                    text=f"*{'I found some terminology' if len(whatises) > 1 else 'I found a whatis'} matching :* {original_query} :muscle:"
                ),
                blocks.DividerBlock(),
            ]
        )
        for wi in whatises:
            block_list.extend(build_whatis_component(wi, is_admin=is_admin))
    else:
        block_list.append(
            blocks.SectionBlock(
                text=f"*No Whatis was found matching the query {original_query}* :shrug:"
                f"\nMaybe it's your time to shine by creating a Whatis of your own!"
            )
        )

    block_list.append(build_whatis_footer(is_admin=is_admin))
    return messages.Message(blocks=block_list, text="Whatis results!")


def build_channel_whatis(triggering_user: str, whatis: Whatis) -> messages.Message:
    """
    Builds a single whatis message to send to channel

    Args:
        triggering_user: The Slack ID of the user who triggered the send to channel action
        whatis: A whatis model object

    Returns: An in-channel whatis message

    """
    channel_blocks = build_whatis_component(whatis, include_buttons=False)
    channel_blocks.insert(
        0,
        blocks.SectionBlock(
            text=f"The user <@{triggering_user}> wants everyone to know about this whatis:"
        ),
    )
    message = messages.Message(text="Whatis channel message", blocks=channel_blocks)
    return message



import typing

from slack.web.classes import dialogs

from whatis import constants
from whatis.models import Whatis

TERMINOLOGY_KEY = "terminology"
DEFINITION_KEY = "definition"
NOTES_KEY = "notes"
LINKS_KEY = "links"
POINT_OF_CONTACT_KEY = "pointofcontact"


def build_create_new_dialog(
    state: typing.Union[dict, str] = ""
) -> dialogs.DialogBuilder:
    d = (
        dialogs.DialogBuilder()
        .submit_label("Add")
        .notify_on_cancel(False)
        .callback_id(constants.CREATE_NEW_WHATIS_ID)
        .title("Create a new Whatis")
        .state(state)
        .text_field(
            label="Terminology",
            name=TERMINOLOGY_KEY,
            optional=False,
            hint="what are you defining",
            min_length=2,
        )
        .text_field(
            label="Definition",
            name=DEFINITION_KEY,
            optional=False,
            hint="What does it mean",
            min_length=8,
        )
        .text_area(
            label="Notes",
            name=NOTES_KEY,
            optional=True,
            hint="Other notes",
            placeholder="Is there anything else you think your teammates should know about this terminology?",
            min_length=8,
        )
        .text_area(
            label="Relevant Links",
            name=LINKS_KEY,
            optional=True,
            hint="Comma separated links to documentation",
            placeholder="Did you know that you can create hyperlinks by <www.thingtolinkto.com | "
            "writing it like this>! Useful for long links",
            min_length=8,
        )
        .conversation_selector(
            label="Point of Contact",
            name=POINT_OF_CONTACT_KEY,
            optional=True,
            placeholder="Where is the best place to ask about this",
        )
    )
    return d


def build_update_dialog(whatis: Whatis) -> dialogs.DialogBuilder:
    d = (
        dialogs.DialogBuilder()
        .submit_label("Add")
        .notify_on_cancel(False)
        .callback_id(constants.UPDATE_WHATIS_ID)
        .title("Update Whatis")
        .state({"id": whatis.id})
        .text_field(
            label="Terminology",
            name=TERMINOLOGY_KEY,
            optional=False,
            hint="what are you defining",
            min_length=2,
            value=whatis.terminology,
        )
        .text_field(
            label="Definition",
            name=DEFINITION_KEY,
            optional=False,
            hint="What does it mean",
            min_length=8,
            value=whatis.definition,
        )
        .text_area(
            label="Notes",
            name=NOTES_KEY,
            optional=True,
            hint="Other notes",
            min_length=8,
            value=whatis.notes,
        )
        .text_area(
            label="Relevant Links",
            name=LINKS_KEY,
            optional=True,
            hint="Comma separated links to documentation",
            min_length=8,
            value=whatis.links,
        )
        .conversation_selector(
            label="Point of Contact",
            name=POINT_OF_CONTACT_KEY,
            optional=True,
            value=whatis.point_of_contact,
            placeholder=whatis.added_by,
        )
    )
    return d



"""create tables

Revision ID: 02809806f723
Revises: 
Create Date: 2019-10-05 00:17:07.343363

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "02809806f723"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "whatis",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("whatis_id", sa.String(), nullable=False),
        sa.Column("terminology", sa.String(), nullable=False),
        sa.Column("definition", sa.String(), nullable=False),
        sa.Column("notes", sa.String(), nullable=True),
        sa.Column("links", sa.String(), nullable=True),
        sa.Column("submitted_at", sa.DateTime(), nullable=False),
        sa.Column("version", sa.Integer(), nullable=False),
        sa.Column("added_by", sa.String(), nullable=False),
        sa.Column("point_of_contact", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "whatis_preloader",
        sa.Column("hash", sa.String(), nullable=False),
        sa.Column("filename", sa.String(), nullable=False),
        sa.Column("loaded_at", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("hash"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("whatis_preloader")
    op.drop_table("whatis")
    # ### end Alembic commands ###



from collections import defaultdict


class SlackInteractionNotRegistered(Exception):
    ...


class SlackInteractionHandler:
    def __init__(self):
        self._interactions = defaultdict(dict)

    def add_interaction(self, interaction_id, f):

        self._interactions[interaction_id] = f

    def interaction(self, interaction_id):
        def wrap(f):
            self.add_interaction(interaction_id, f)
            return f

        return wrap

    def interact(self, interaction_id, **kwargs):
        r = self._interactions.get(interaction_id)

        if r is None:
            raise SlackInteractionNotRegistered(
                f"No function registered for the interaction ID {interaction_id}"
            )
        else:
            return r(**kwargs)



from flask import current_app
from slack.web.classes.interactions import DialogInteractiveEvent

from whatis import constants
from whatis.routes.actions import create_whatis, update_whatis
from whatis.utils.dialog_components import (
    LINKS_KEY,
    NOTES_KEY,
    TERMINOLOGY_KEY,
    DEFINITION_KEY,
    POINT_OF_CONTACT_KEY,
)
from whatis.utils.interaction_handler import SlackInteractionHandler
from whatis.utils.responder import webhook_response, basic_responder_response

dialog_interactor = SlackInteractionHandler()


def handle_dialog_submission(data: dict):
    ds = DialogInteractiveEvent(data)
    result = dialog_interactor.interact(ds.callback_id, data=ds)
    wr = webhook_response(ds.response_url, result)
    current_app.logger.info(
        f"Dialog sumbit interaction for {ds.callback_id} returning {result} - {wr}"
    )
    return "", 204


@dialog_interactor.interaction(constants.CREATE_NEW_WHATIS_ID_SUBMIT)
def create_dialog_sumbit_action(data: DialogInteractiveEvent):
    submission = data.submission
    whatis = create_whatis(
        terminology=submission.get(TERMINOLOGY_KEY),
        definition=submission.get(DEFINITION_KEY),
        notes=submission.get(NOTES_KEY),
        links=submission.get(LINKS_KEY),
        version=1,
        added_by=data.user.id,
        point_of_contact=submission.get(POINT_OF_CONTACT_KEY),
    )
    return basic_responder_response(
        f"Successfully created the Whatis {whatis.terminology} :tada:"
    ).to_dict()


@dialog_interactor.interaction(constants.UPDATE_WHATIS_ID_SUBMIT)
def update_dialog_sumbit_action(data: DialogInteractiveEvent):
    submission = data.submission
    id = data.state["id"]
    whatis = update_whatis(
        id=int(id),
        definition=submission.get(DEFINITION_KEY),
        notes=submission.get(NOTES_KEY),
        links=submission.get(LINKS_KEY),
        added_by=data.user.id,
        point_of_contact=submission.get(POINT_OF_CONTACT_KEY),
    )
    return basic_responder_response(
        f"Successfully updated the Whatis {whatis.terminology} :tada:"
    ).to_dict()



from datetime import datetime
from functools import partial
from secrets import token_urlsafe

from flask_sqlalchemy import SQLAlchemy

db = SQLAlchemy()

# whatis_id generator
widgen = partial(token_urlsafe, 5)


class Whatis(db.Model):  # type: ignore
    id = db.Column(db.Integer, primary_key=True)
    whatis_id = db.Column(db.String, nullable=False, default=widgen)
    terminology = db.Column(db.String, nullable=False)
    definition = db.Column(db.String, nullable=False)
    notes = db.Column(db.String, nullable=True)
    links = db.Column(db.String, nullable=True)
    submitted_at = db.Column(db.DateTime, default=datetime.now, nullable=False)
    version = db.Column(db.Integer, nullable=False)
    added_by = db.Column(db.String, nullable=False)
    point_of_contact = db.Column(db.String, nullable=True)

    def __repr__(self):
        return f"<Whatis {self.terminology} - {self.id}>"


class WhatisPreloader(db.Model):  # type: ignore
    """
    A simple table to record whether or not a Terminology CSV has already been loaded
    """

    hash = db.Column(db.String, primary_key=True)
    filename = db.Column(db.String, nullable=False)
    loaded_at = db.Column(db.DateTime, nullable=False, default=datetime.now)

    def __repr__(self):
        return f"<WhatisPreload for file {self.filename} at {self.loaded_at}>"



from setuptools import setup, find_packages
from whatis import __version__

# Get requirements for dependencies
requirements = open("requirements.txt").read().split("\n")

setup(
    name="whatis-bot",
    version=__version__,
    packages=find_packages(),
    url="https://github.com/wooddar/whatis",
    license="mit",
    author="Hugo Darwood",
    scripts=["scripts/whatis-serve"],
    author_email="hugodarwood@gmail.com",
    description="A Slack app to explore, create and curate your workspace's business terminology",
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)



# Interaction IDS
CREATE_NEW_WHATIS_ID = "whatis-create-new"
CREATE_NEW_WHATIS_ID_SUBMIT = "whatis-create-new"
UPDATE_WHATIS_ID = "whatis-update"
UPDATE_WHATIS_ID_SUBMIT = "whatis-update"
DELETE_WHATIS_ID = "whatis-delete"
ROLLBACK_WHATIS_ID = "whatis-rollback"
WHATIS_ALL_ID = "whatis-all"
WHATIS_SEND_CHANNEL_ID = "whatis-send-channel"

# Whatis visible fields
WHATIS_FIELDS = [
    "terminology",
    "definition",
    "notes",
    "links",
    "point_of_contact",
    "added_by",
]



import requests
from slack.web.classes import messages


def webhook_response(response_url: str, json=None):
    return requests.post(response_url, json=json)


def basic_responder_response(text: str) -> messages.Message:
    return messages.Message(text=text)



import json
import logging
import os
import uuid
from hashlib import sha224
from pathlib import Path

from alembic import command
from alembic.migration import MigrationContext
from cachetools import cached, TTLCache
from flask import Flask
from flask_migrate import Migrate
from slack.errors import SlackApiError
from slack.web.client import WebClient

from .constants import WHATIS_FIELDS
from .models import db as sqlalchemy_db, WhatisPreloader, Whatis

logging.basicConfig(level=logging.DEBUG)


class WhatisApp(Flask):
    def __init__(self, db_uri=None, config=None, preload_path=None, **kwargs):
        Flask.__init__(self, __name__)

        # Preload default configuration
        self.config.from_object(config)
        self.config.from_mapping(kwargs)

        # Set the secret key for this instance (creating one if one does not exist already)
        self.config["SECRET_KEY"] = self.config["SECRET_KEY"] or str(uuid.uuid4())

        # Configure database
        if db_uri:
            self.config["SQLALCHEMY_DATABASE_URI"] = db_uri
        if self.config["SQLALCHEMY_DATABASE_URI"] == "sqlite:///:memory:":
            self.logger.warning(
                "Using Sqlite in-memory database, all data will be lost when server shuts down!"
            )

        # DB dialect logic - used for lookup operations
        db_dialect = self.config["SQLALCHEMY_DATABASE_URI"].split(":")[0]
        self.logger.info(f"Attempting to use db dialect {db_dialect}")
        if self.config.get("DEBUG") is not True:
            self.logger.warning(
                "It is strongly recommended that you do not use Sqlite for production deployments!"
            )
        if not any([i == db_dialect for i in ["postgres", "sqlite"]]):
            raise RuntimeError(
                f"Dialect {db_dialect} not supported - please use sqlite or postgres"
            )
        self.config["DB_DIALECT"] = db_dialect

        # Register database schema with flask app
        sqlalchemy_db.init_app(self)

        # Set up database migration information
        # Registers Migrate plugin in self.extensions['migrate']
        Migrate(self, self.db)

        # Try to create the database if it does not already exist
        # Existence is determined by whether there is an existing alembic migration revision
        db_auto_create = self.config.get("DB_AUTO_CREATE", True)
        db_auto_upgrade = self.config.get("DB_AUTO_UPGRADE", True)
        if db_auto_create and self.db_revision is None:
            self.db_init()
        elif db_auto_upgrade:
            self.db_upgrade()

        self.logger.setLevel(logging.DEBUG)

        # Install postgres fuzzystrmatch extension
        if db_dialect == "postgres":
            self.logger.info("Enabling Postgres fuzzy string matching")
            with self.app_context(), self.db.engine.connect() as conn:
                conn.execute("CREATE EXTENSION IF NOT EXISTS  fuzzystrmatch")

        # Handle preloading an existing Terminology set
        self.handle_whatis_preload(preload_path)

        # Register Slack client on the current application instance
        if all(
            [
                self.config.get(i) is None
                for i in ["SLACK_SIGNING_SECRET", "SLACK_TOKEN"]
            ]
        ):
            raise RuntimeError(
                "Whatis must have both a slack signing secret and slack bot token set"
            )
        self.sc = WebClient(self.config.get("SLACK_TOKEN"), ssl=False)

        from whatis.routes.slack_route import slack_blueprint

        self.register_blueprint(slack_blueprint, url_prefix="/slack")

        if not all(
            [
                type(self.config[i]) == list
                for i in ["ADMIN_USER_IDS", "ADMIN_CHANNEL_IDS"]
            ]
        ):
            raise RuntimeError(
                "ADMIN_USER_IDS and ADMIN_CHANNEL_IDS must be lists of Admin user IDs or channel IDs"
            )

        try:
            au = self.admin_users
            self.logger.info(f"Initial Admin users set as {au}")
        except SlackApiError as s:
            raise RuntimeError(
                f"Failed to get Admin users from specified Admin channels - are you sure the whatis bot "
                f"is invited and has the necessary scopes {s}"
            )

        # Register a basic route for healthchecking
        @self.route("/ping")
        def healthcheck():
            return "pong"

    @property
    def db(self):
        return sqlalchemy_db

    @property
    def _alembic_config(self):
        if not hasattr(self, "extensions") or "migrate" not in self.extensions:
            raise RuntimeError(
                "KnowledgeApp has not yet been configured. Please instantiate it via `get_app_for_repo`."
            )
        migrations_path = os.path.join(os.path.dirname(__file__), "migrations")
        # This is terrible but seems to be needed for packaging
        migrations_path = (
            migrations_path
            if Path(migrations_path).exists() is True
            else os.path.join(os.path.dirname(__file__), "whatis/migrations")
        )
        return self.extensions["migrate"].migrate.get_config(migrations_path)

    def db_init(self):
        with self.app_context():
            # Create all tables
            sqlalchemy_db.create_all()

            # Stamp table as being current
            command.stamp(self._alembic_config, "head")
        return self

    @property
    def db_revision(self):
        with self.app_context():
            conn = self.db.engine.connect()

            context = MigrationContext.configure(conn)
            return context.get_current_revision()

    @cached(TTLCache(ttl=3600, maxsize=2048))
    def _get_admin_users(self):
        channel_admin_members = []
        for channel in self.config["ADMIN_CHANNEL_IDS"]:
            try:
                channel_admin_members.extend(
                    self.sc.conversations_members(channel=channel)["members"]
                )
            except SlackApiError as s:
                self.logger.warning(
                    f"Could not get members from the specified Admin channel {channel} has the whatis "
                    f"bot been removed or scopes been changed? {s}"
                )
        return self.config["ADMIN_USER_IDS"] + channel_admin_members

    @property
    def admin_users(self):
        """
        Get all users approved as admins
        """
        return self._get_admin_users()

    def db_upgrade(self):
        with self.app_context():
            command.upgrade(self._alembic_config, "head")
        return self

    def db_downgrade(self, revision):
        with self.app_context():
            command.downgrade(self._alembic_config, revision)
        return self

    def db_migrate(self, message, autogenerate=True):
        with self.app_context():
            command.revision(
                self._alembic_config, message=message, autogenerate=autogenerate
            )
        return self

    def handle_whatis_preload(self, preload_path):
        with self.app_context():
            if preload_path is not None:
                filepath = Path(preload_path)
                if filepath.exists():
                    file_contents = open(filepath).read()
                    file_hash = sha224(file_contents.encode()).hexdigest()
                    existing_preload = (
                        self.db.session.query(WhatisPreloader)
                        .filter(WhatisPreloader.hash == file_hash)
                        .first()
                    )
                    if existing_preload is not None:
                        self.logger.info(
                            f"Existing whatis load found for the file {existing_preload}, skipping"
                        )
                    else:
                        raw = json.loads(file_contents)
                        self.logger.info(
                            f"Attempting to preload terminology from the file {filepath}, {len(raw)} records found"
                        )
                        for raw_whatis in raw:
                            if all(
                                [i in raw_whatis for i in ["terminology", "definition"]]
                            ) is False or not set(raw_whatis).issubset(
                                set(WHATIS_FIELDS)
                            ):
                                raise RuntimeError(
                                    f"Attempt to preload Terminology failed, Whatis {raw_whatis} has an unrecognised attribute or does not contain both of [terminology, definiton]"
                                )
                            wi = Whatis(
                                **{"version": 0, "added_by": "WHATIS BOT", **raw_whatis}
                            )
                            self.db.session.add(wi)
                            self.logger.debug(f"Added the Whatis {wi}")
                        # Now register the fact that we have loaded this file so it is ignored for future deployments
                        self.db.session.add(
                            WhatisPreloader(hash=file_hash, filename=str(filepath))
                        )
                        self.db.session.commit()

                else:
                    raise FileNotFoundError(
                        f"Preload filepath specified {preload_path} but no file found"
                    )
